from .utils import clean_setting


# minimum standing a character needs to have in order to get alliance contacts
# Any char with a standing smaller than this value will be rejected
STANDINGSSYNC_CHAR_MIN_STANDING = \
    clean_setting(
        'STANDINGSSYNC_CHAR_MIN_STANDING', 
        default_value=0.1,
        min_value=-10, 
        max_value=10
    )



import logging

from django.db import models
from .utils import LoggerAddTag


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


class AllianceContactManager(models.Manager):    
    
    def grouped_by_standing(self, sync_manager: object) -> dict:
        """returns alliance contacts grouped by their standing as dict"""
        from .models import AllianceContact

        contacts = AllianceContact.objects.filter(manager=sync_manager)
        contacts_by_standing = dict()
        for contact in contacts:
            standing = contact.standing
            if standing not in contacts_by_standing:
                contacts_by_standing[standing] = set()
            
            contacts_by_standing[standing].add(contact)
        
        return contacts_by_standing



import logging

from django.db import models
from django.utils.timezone import now

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveAllianceInfo, EveCharacter

from .managers import AllianceContactManager
from .utils import LoggerAddTag

logger = LoggerAddTag(logging.getLogger(__name__), __package__)


class SyncManager(models.Model):
    """An object for managing syncing of contacts for an alliance"""

    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_INSUFFICIENT_PERMISSIONS = 3    
    ERROR_NO_CHARACTER = 4
    ERROR_ESI_UNAVAILABLE = 5
    ERROR_UNKNOWN = 99

    ERRORS_LIST = [
        (ERROR_NONE, 'No error'),
        (ERROR_TOKEN_INVALID, 'Invalid token'),
        (ERROR_TOKEN_EXPIRED, 'Expired token'),
        (ERROR_INSUFFICIENT_PERMISSIONS, 'Insufficient permissions'),
        (ERROR_NO_CHARACTER, 'No character set for fetching alliance contacts'),
        (ERROR_ESI_UNAVAILABLE, 'ESI API is currently unavailable'),
        (ERROR_UNKNOWN, 'Unknown error'),
    ]

    alliance = models.OneToOneField(
        EveAllianceInfo, 
        on_delete=models.CASCADE,
        primary_key=True
    )
    # alliance contacts are fetched from this character
    character = models.OneToOneField(
        CharacterOwnership, 
        on_delete=models.SET_NULL, 
        null=True, 
        default=None
    )
    version_hash = models.CharField(max_length=32, null=True, default=None)    
    last_sync = models.DateTimeField(null=True, default=None)
    last_error = models.IntegerField(choices=ERRORS_LIST, default=ERROR_NONE)
    
    def __str__(self):
        if self.character is not None:
            character_name = self.character.character.character_name
        else:
            character_name = 'None'
        return '{} ({})'.format(
            self.alliance.alliance_name, 
            character_name
        )

    def set_sync_status(self, status):
        """sets the sync status with the current date and time"""
        self.last_error = status
        self.last_sync = now()
        self.save()

    def get_effective_standing(self, character: EveCharacter):
        """ return the effective standing with this alliance"""
        
        contacts = AllianceContact.objects.filter(manager=self)

        # check if character is in contacts
        c = [
            x for x in contacts 
            if x.contact_id == int(character.character_id)
        ]
        # else check if character's corporation is in contacts
        if not c:
            c = [
                x for x in contacts 
                if x.contact_id == int(character.corporation_id)
            ]
            # else check if character's alliances is in contacts
            if character.alliance_id is not None and not c:
                c = [
                    x for x in contacts 
                    if x.contact_id == int(character.alliance_id)
                ]
        
        return c.pop().standing if c else 0
    
    @classmethod
    def get_esi_scopes(cls) -> list:
        return ['esi-alliances.read_contacts.v1']


class SyncedCharacter(models.Model):    
    """A character that has his personal contacts synced with an alliance"""
    
    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_INSUFFICIENT_PERMISSIONS = 3
    ERROR_ESI_UNAVAILABLE = 5
    ERROR_UNKNOWN = 99

    ERRORS_LIST = [
        (ERROR_NONE, 'No error'),
        (ERROR_TOKEN_INVALID, 'Invalid token'),
        (ERROR_TOKEN_EXPIRED, 'Expired token'),
        (ERROR_INSUFFICIENT_PERMISSIONS, 'Insufficient permissions'),
        (ERROR_ESI_UNAVAILABLE, 'ESI API is currently unavailable'),
        (ERROR_UNKNOWN, 'Unknown error'),
    ]
        
    character = models.OneToOneField(
        CharacterOwnership, 
        on_delete=models.CASCADE,
        primary_key=True
    )
    manager = models.ForeignKey(SyncManager, on_delete=models.CASCADE)
    version_hash = models.CharField(max_length=32, null=True, default=None)    
    last_sync = models.DateTimeField(null=True, default=None)
    last_error = models.IntegerField(choices=ERRORS_LIST, default=ERROR_NONE)
    
    def __str__(self):
        return self.character.character.character_name

    def set_sync_status(self, status):
        """sets the sync status with the current date and time"""
        self.last_error = status
        self.last_sync = now()
        self.save()

    def get_status_message(self):
        if self.last_error != self.ERROR_NONE:
            message = self.get_last_error_display()
        elif self.last_sync is not None:            
            message = 'OK'
        else:
            message = 'Not synced yet'
        
        return message

    @staticmethod
    def get_esi_scopes() -> list:
        return [
            'esi-characters.read_contacts.v1', 
            'esi-characters.write_contacts.v1'
        ]
    

class AllianceContact(models.Model):
    """An alliance contact with standing"""    
    manager = models.ForeignKey(SyncManager, on_delete=models.CASCADE)
    contact_id = models.IntegerField()
    contact_type = models.CharField(max_length=32)
    standing = models.FloatField()

    objects = AllianceContactManager()

    def __str__(self):
        return '{}:{}'.format(self.contact_type, self.contact_id)
        
    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['manager', 'contact_id'], 
                name="manager-contacts-unq")
        ]



from django.apps import AppConfig
from . import __version__


class SyncAltContactsConfig(AppConfig):
    name = 'standingssync'
    label = 'standingssync'
    verbose_name = 'Standings Sync v{}'.format(__version__)



import hashlib
import json
import logging

from celery import shared_task

from django.db import transaction
from django.contrib.auth.models import User

from allianceauth.notifications import notify
from esi.models import Token
from esi.errors import TokenExpiredError, TokenInvalidError

from .app_settings import STANDINGSSYNC_CHAR_MIN_STANDING
from .helpers.esi_fetch import esi_fetch
from .models import SyncManager, SyncedCharacter, AllianceContact
from .utils import LoggerAddTag, make_logger_prefix, chunks


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


@shared_task
def run_character_sync(sync_char_pk, force_sync=False, manager_pk=None):
    """syncs contacts for one character

    Will delete the sync character if necessary,
    e.g. if token is no longer valid or character is no longer blue
    
    Args:
        sync_char_pk: primary key of sync character to run sync for
        force_sync: will ignore version_hash if set to true
        mananger_pk: when provided will override related sync manager

    Returns:
        False if the sync character was deleted, True otherwise
    """

    def issue_message(synced_character, message): 
        return (
            'Standings Sync has been deactivated for your '
            'character {}, because {}.\n'
            'Feel free to activate sync for your character again, '
            'once the issue has been resolved.'.format(
                synced_character, message
            )
        )

    try:
        synced_character = SyncedCharacter.objects.get(pk=sync_char_pk)
    except SyncedCharacter.DoesNotExist:
        raise SyncedCharacter.DoesNotExist(
            "Requested character with pk {} does not exist".format(sync_char_pk)
        )
    addTag = make_logger_prefix(synced_character)
    
    user = synced_character.character.user
    issue_title = 'Standings Sync deactivated for {}'.format(synced_character)
    
    # abort if owner does not have sufficient permissions
    if not user.has_perm('standingssync.add_syncedcharacter'):        
        logger.info(addTag(
            'sync deactivated due to insufficient user permissions'
        ))
        notify(
            user, 
            issue_title, 
            issue_message(
                synced_character, 
                'you no longer have permission for this service'
            )
        )
        synced_character.delete()
        return False

    # check if an update is needed
    if manager_pk is None:
        manager = synced_character.manager
    else:
        manager = SyncManager.objects.get(pk=manager_pk)

    if (
        not force_sync 
        and manager.version_hash == synced_character.version_hash
    ):
        logger.info(addTag(
            'contacts of this char are up-to-date, no sync required'
        ))
    else:        
        # get token
        try:
            token = Token.objects\
                .filter(
                    user=synced_character.character.user, 
                    character_id=synced_character.character.character.character_id
                )\
                .require_scopes(SyncedCharacter.get_esi_scopes())\
                .require_valid().first()

        except TokenInvalidError:
            logger.info(addTag('sync deactivated due to invalid token'))
            notify(
                user, 
                issue_title, 
                issue_message(
                    synced_character,
                    'your token is no longer valid'
                )
            )
            synced_character.delete()
            return False

        except TokenExpiredError:
            logger.info(addTag('sync deactivated due to expired token'))
            notify(
                user, 
                issue_title, 
                issue_message(synced_character, 'your token has expired')
            )
            synced_character.delete()
            return False

        if (
            manager.get_effective_standing(
                synced_character.character.character
            ) < STANDINGSSYNC_CHAR_MIN_STANDING
        ):
            notify(
                user, 
                issue_title, 
                issue_message(
                    synced_character,
                    'your character is no longer blue with the alliance'
                )
            )
            synced_character.delete()
            return False
        
        if token is None:            
            synced_character.set_sync_status(SyncedCharacter.ERROR_UNKNOWN)
            raise RuntimeError('Can not find suitable token for synced char')
        
        try:
            _perform_contacts_sync_for_character(
                synced_character, token, addTag
            )
            
        except Exception as ex:
            logger.error('An unexpected error ocurred: {}'.format(ex))            
            synced_character.set_sync_status(SyncedCharacter.ERROR_UNKNOWN)
            raise

    return True
    

def _perform_contacts_sync_for_character(synced_character, token, addTag):    
    logger.info(addTag('Updating contacts with new version'))    
    character_id = synced_character.character.character.character_id
    # get current contacts
    character_contacts = esi_fetch(
        'Contacts.get_characters_character_id_contacts',
        args={'character_id': character_id},
        has_pages=True,
        token=token
    )    
    # delete all current contacts
    max_items = 20
    contact_ids_chunks = chunks(
        [x['contact_id'] for x in character_contacts], max_items
    )
    for contact_ids_chunk in contact_ids_chunks:
        esi_fetch(
            'Contacts.delete_characters_character_id_contacts',
            args={
                'character_id': character_id,
                'contact_ids': contact_ids_chunk
            }, 
            token=token
        )
    
    # write alliance contacts to ESI
    contacts_by_standing = AllianceContact.objects\
        .grouped_by_standing(sync_manager=synced_character.manager)
    max_items = 100
    for standing in contacts_by_standing.keys():                
        contact_ids_chunks = chunks(
            [c.contact_id for c in contacts_by_standing[standing]], max_items
        )
        for contact_ids_chunk in contact_ids_chunks:
            esi_fetch(
                'Contacts.post_characters_character_id_contacts',
                args={
                    'character_id': character_id,
                    'contact_ids': contact_ids_chunk,
                    'standing': standing
                }, 
                token=token
            )

    # store updated version hash with character
    synced_character.version_hash = synced_character.manager.version_hash
    synced_character.save()
    synced_character.set_sync_status(SyncedCharacter.ERROR_NONE)


@shared_task
def run_manager_sync(manager_pk, force_sync=False, user_pk=None):
    """sync contacts and related characters for one manager

    Args:
        manage_pk: primary key of sync manager to run sync for
        force_sync: will ignore version_hash if set to true
        user_pk: user to send a completion report to (optional)

    Returns:
        True on success or False on error
    """
    
    try:
        sync_manager = SyncManager.objects.get(pk=manager_pk)
    except SyncManager.DoesNotExist:        
        raise SyncManager.DoesNotExist(
            'task called for non existing manager with pk {}'.format(manager_pk)
        )        
    
    try:
        addTag = make_logger_prefix(sync_manager)
        
        if sync_manager.character is None:
            logger.error(addTag(
                'No character configured to sync the alliance'
            ))
            sync_manager.set_sync_status(SyncManager.ERROR_NO_CHARACTER)
            raise ValueError()
        
        # abort if character does not have sufficient permissions
        if not sync_manager.character.user.has_perm(
            'standingssync.add_syncmanager'
        ):
            logger.error(addTag(
                'Character does not have sufficient permission '
                'to sync the alliance'
            ))
            sync_manager.set_sync_status(
                SyncManager.ERROR_INSUFFICIENT_PERMISSIONS
            )
            raise ValueError()

        try:            
            # get token    
            token = Token.objects\
                .filter(
                    user=sync_manager.character.user, 
                    character_id=sync_manager.character.character.character_id
                )\
                .require_scopes(SyncManager.get_esi_scopes())\
                .require_valid().first()

        except TokenInvalidError:        
            logger.error(addTag(
                'Invalid token for fetching alliance contacts'
            ))
            sync_manager.set_sync_status(SyncManager.ERROR_TOKEN_INVALID)
            raise TokenInvalidError()
            
        except TokenExpiredError:
            sync_manager.set_sync_status(SyncManager.ERROR_TOKEN_EXPIRED)
            raise TokenExpiredError()

        else:
            if not token:
                sync_manager.set_sync_status(SyncManager.ERROR_TOKEN_INVALID)
                raise TokenInvalidError()
        
        try:
            _perform_contacts_sync_for_manager(
                sync_manager, token, addTag, force_sync
            )
            
        except Exception as ex:
            logger.error(addTag(
                'An unexpected error ocurred while trying to '
                'sync alliance: {}'. format(ex)
            ))
            sync_manager.set_sync_status(SyncManager.ERROR_UNKNOWN)
            raise RuntimeError()

    except Exception:
        success = False        
    else:
        success = True
        
    if user_pk:
        try:
            message = 'Syncing of alliance contacts for "{}" {}.\n'.format(
                sync_manager.alliance,
                'completed successfully' if success else 'has failed'
            )
            if success:
                message += '{:,} contacts synced.'.format(
                    sync_manager.alliancecontact_set.count()
                )
            
            notify(
                user=User.objects.get(pk=user_pk),
                title='Standings Sync: Alliance sync for {}: {}'.format(
                    sync_manager.alliance,
                    'OK' if success else 'FAILED'
                ),
                message=message,
                level='success' if success else 'danger'
            )
        except Exception as ex:
            logger.error(addTag(
                'An unexpected error ocurred while trying to '
                'report to user: {}'. format(ex)
            ))
    
    return success


def _perform_contacts_sync_for_manager(
    sync_manager, token, addTag, force_sync
):        
    # get alliance contacts 
    alliance_id = int(sync_manager.character.character.alliance_id)
    contacts = esi_fetch(
        'Contacts.get_alliances_alliance_id_contacts',
        args={'alliance_id': alliance_id},
        has_pages=True,
        token=token
    ) 

    # determine if contacts have changed by comparing their hashes
    new_version_hash = hashlib.md5(
        json.dumps(contacts).encode('utf-8')
    ).hexdigest()
    if force_sync or new_version_hash != sync_manager.version_hash:
        logger.info(addTag(
            'Storing alliance update with {:,} contacts'.format(
                len(contacts)
            ))
        )                
        contacts_unique = {int(c['contact_id']): c for c in contacts}
        
        # add the sync alliance with max standing to contacts
        contacts_unique[alliance_id] = {
            'contact_id': alliance_id,
            'contact_type': 'alliance',
            'standing': 10
        }        
        with transaction.atomic():
            AllianceContact.objects\
                .filter(manager=sync_manager)\
                .delete()
            for contact_id, contact in contacts_unique.items():
                AllianceContact.objects.create(
                    manager=sync_manager,
                    contact_id=contact_id,
                    contact_type=contact['contact_type'],
                    standing=contact['standing']                        
                )
            sync_manager.version_hash = new_version_hash
            sync_manager.save()
    else:
        logger.info(addTag('Alliance contacts are unchanged.'))
    
    # dispatch tasks for characters that need syncing
    alts_need_syncing = SyncedCharacter.objects\
        .filter(manager=sync_manager)\
        .exclude(version_hash=new_version_hash)
    for character in alts_need_syncing:
        run_character_sync.delay(character.pk)

    sync_manager.set_sync_status(SyncManager.ERROR_NONE)
    

@shared_task
def run_regular_sync():
    """syncs all managers and related characters if needed"""        
    for sync_manager in SyncManager.objects.all():
        run_manager_sync.delay(sync_manager.pk)



import logging

from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required, permission_required

from esi.decorators import token_required
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveAllianceInfo

from . import tasks, __title__
from .app_settings import STANDINGSSYNC_CHAR_MIN_STANDING
from .models import SyncManager, SyncedCharacter, AllianceContact
from .utils import LoggerAddTag, messages_plus


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


@login_required
@permission_required('standingssync.add_syncedcharacter')
def index(request):        
    """main page"""
    if request.user.profile.main_character is None:
        sync_manager = None    
    else:
        try:
            alliance = EveAllianceInfo.objects.get(
                alliance_id=request.user.profile.main_character.alliance_id
            )
            sync_manager = SyncManager.objects.get(alliance=alliance)
        except EveAllianceInfo.DoesNotExist:
            sync_manager = None
        except SyncManager.DoesNotExist:
            sync_manager = None

    # get list of synced characters for this user
    characters_query = SyncedCharacter.objects\
        .select_related('character__character')\
        .filter(character__user=request.user)

    characters = list()
    for character in characters_query:                        
        characters.append({
            'portrait_url': character.character.character.portrait_url,
            'name': character.character.character.character_name,
            'status_message': character.get_status_message(),
            'has_error': character.last_error != SyncedCharacter.ERROR_NONE,
            'pk': character.pk
        })
    
    has_synced_chars = characters_query.count() > 0
    context = {        
        'app_title': __title__,
        'characters': characters,
        'has_synced_chars': has_synced_chars        
    }
    if sync_manager is not None:
        context['alliance'] = sync_manager.alliance
        context['alliance_contacts_count'] = AllianceContact.objects\
            .filter(manager=sync_manager)\
            .count()
    else:
        context['alliance'] = None
        context['alliance_contacts_count'] = None

    return render(request, 'standingssync/index.html', context)


@login_required
@permission_required('standingssync.add_syncmanager')
@token_required(SyncManager.get_esi_scopes())
def add_alliance_manager(request, token):
    """add or update sync manager for an alliance"""
    success = True
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    if not token_char.alliance_id:
        messages_plus.warning(
            request, 
            (
                'Can not add {}, because it is not a member '
                'of any alliance. '.format(token_char)
            )
        )
        success = False
    
    if success:
        try:
            owned_char = CharacterOwnership.objects.get(
                user=request.user,
                character=token_char
            )            
        except CharacterOwnership.DoesNotExist:
            messages_plus.warning(
                request, 'Could not find character {}'.format(
                    token_char.character_name
                )    
            )
            success = False
    
    if success:
        try:
            alliance = EveAllianceInfo.objects.get(
                alliance_id=token_char.alliance_id
            )
        except EveAllianceInfo.DoesNotExist:
            alliance = EveAllianceInfo.objects.create_alliance(
                token_char.alliance_id
            )
            alliance.save()

    if success:
        sync_manager, _ = SyncManager.objects.update_or_create(    
            alliance=alliance,
            defaults={
                'character': owned_char
            }
        )  
        tasks.run_manager_sync.delay(
            manager_pk=sync_manager.pk, user_pk=request.user.pk
        )
        messages_plus.success(
            request, 
            '{} set as alliance character for {}. '
            'Started syncing of alliance contacts. '
            'You will receive a report once it is completed.'.format(
                sync_manager.character.character.character_name, 
                alliance.alliance_name
            )
        )
    return redirect('standingssync:index')


@login_required
@permission_required('standingssync.add_syncedcharacter')
@token_required(scopes=SyncedCharacter.get_esi_scopes())
def add_character(request, token):
    """add character to receive alliance contacts"""
    try:        
        alliance = EveAllianceInfo.objects.get(
            alliance_id=request.user.profile.main_character.alliance_id
        )
    except EveAllianceInfo.DoesNotExist:
        raise RuntimeError("Can not find alliance")
    
    try:
        sync_manager = SyncManager.objects.get(alliance=alliance)
    except SyncManager.DoesNotExist:
        raise RuntimeError("can not find sync manager for alliance")
        
    token_char = EveCharacter.objects.get(character_id=token.character_id)
    if token_char.alliance_id == sync_manager.character.character.alliance_id:
        messages_plus.warning(
            request,
            'Adding alliance members does not make much sense, '
            'since they already have access to alliance contacts.'
        )

    else:
        try:
            owned_char = CharacterOwnership.objects.get(
                user=request.user, character=token_char
            )
        except CharacterOwnership.DoesNotExist:
            messages_plus.warning(
                request,
                'Could not find character {}'.format(token_char.character_name)    
            )
        else:
            if (
                sync_manager.get_effective_standing(owned_char.character) 
                < STANDINGSSYNC_CHAR_MIN_STANDING
            ):
                messages_plus.warning(
                    request,
                    'Can not activate sync for your character {} '
                    ', because it does not have blue standing '
                    'with the alliance. Please first obtain blue '
                    'standing for your character and then try again.'.format(
                        token_char.character_name
                    )
                )
            else:
                sync_character, _ = SyncedCharacter.objects.update_or_create(
                    character=owned_char,
                    defaults={'manager': sync_manager}
                )
                tasks.run_character_sync.delay(sync_character.pk)
                messages_plus.success(
                    request, 
                    'Sync activated for {}!'.format(token_char.character_name)
                )    
    return redirect('standingssync:index')


@login_required
@permission_required('standingssync.add_syncedcharacter')
def remove_character(request, alt_pk):
    """remove character from receiving alliance contacts"""
    alt = SyncedCharacter.objects.get(pk=alt_pk)
    alt_name = alt.character.character.character_name
    alt.delete()
    messages_plus.success(
        request, 
        'Sync deactivated for {}'.format(alt_name)
    )    
    return redirect('standingssync:index')



# Generated by Django 2.2.5 on 2019-09-19 11:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    
    initial = True

    dependencies = [
        ('eveonline', '0010_alliance_ticker'),
        ('authentication', '0016_ownershiprecord'),
    ]

    operations = [
        migrations.CreateModel(
            name='SyncManager',
            fields=[
                ('alliance', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='eveonline.EveAllianceInfo')),
                ('version_hash', models.CharField(default=None, max_length=32, null=True)),
                ('last_sync', models.DateTimeField(default=None, null=True)),
                ('character', models.OneToOneField(default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, to='authentication.CharacterOwnership')),
            ],
        ),
        migrations.CreateModel(
            name='SyncedCharacter',
            fields=[
                ('character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='authentication.CharacterOwnership')),
                ('version_hash', models.CharField(default=None, max_length=32, null=True)),
                ('last_error', models.IntegerField(choices=[(0, 'No error'), (1, 'Invalid token'), (99, 'Unknown error')], default=0)),
                ('last_sync', models.DateTimeField(default=None, null=True)),
                ('manager', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='standingssync.SyncManager')),
            ],
        ),
        migrations.CreateModel(
            name='AllianceContact',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contact_id', models.IntegerField()),
                ('contact_type', models.CharField(max_length=32)),
                ('standing', models.FloatField()),
                ('manager', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='standingssync.SyncManager')),
            ],
        ),
        migrations.AddConstraint(
            model_name='alliancecontact',
            constraint=models.UniqueConstraint(fields=('manager', 'contact_id'), name='manager-contacts-unq'),
        ),
        migrations.AlterField(
            model_name='syncedcharacter',
            name='last_error',
            field=models.IntegerField(choices=[(0, 'No error'), (1, 'Invalid token'), (2, 'Insufficient permissions'), (99, 'Unknown error')], default=0),
        ),
        migrations.AlterField(
            model_name='syncedcharacter',
            name='last_error',
            field=models.IntegerField(choices=[(0, 'No error'), (1, 'Invalid token'), (2, 'Expired token'), (3, 'Insufficient permissions'), (5, 'ESI API is currently unavailable'), (99, 'Unknown error')], default=0),
        ),
        migrations.AddField(
            model_name='syncmanager',
            name='last_error',
            field=models.IntegerField(choices=[(0, 'No error'), (1, 'Invalid token'), (2, 'Expired token'), (3, 'Insufficient permissions'), (4, 'No character set for fetching alliance contacts'), (5, 'ESI API is currently unavailable'), (99, 'Unknown error')], default=0),
        ),
    ]



from django.conf.urls import url
from . import views

app_name = 'standingssync'

urlpatterns = [
    url(r'^$', views.index, name='index'),
    url(r'^add_character/$', views.add_character, name='add_character'),
    url(
        r'^remove_character/(?P<alt_pk>[0-9]+)/$', 
        views.remove_character, 
        name='remove_character'
    ),
    url(
        r'^add_alliance_manager/$', 
        views.add_alliance_manager, 
        name='add_alliance_manager'
    ),    
]



from django.contrib import admin
from .models import SyncedCharacter, SyncManager
from . import tasks


@admin.register(SyncedCharacter)
class SyncedCharacterAdmin(admin.ModelAdmin):
    list_display = (
        'user', 
        'character_name', 
        'version_hash', 
        'last_sync', 
        'last_error', 
        'manager'
    )
    list_filter = (
        'last_error', 
        'version_hash', 
        'last_sync', 
        'character__user', 
        'manager'
    )
    actions = ['start_sync_contacts']
    
    list_display_links = None

    def user(self, obj):
        return obj.character.user

    def character_name(self, obj):
        return obj.__str__()

    # This will help you to disbale add functionality
    def has_add_permission(self, request):
        return False

    def start_sync_contacts(self, request, queryset):
                
        names = list()
        for obj in queryset:            
            tasks.run_character_sync.delay(sync_char_pk=obj.pk, force_sync=True)
            names.append(str(obj))
    
        self.message_user(
            request, 
            'Started syncing for: {}'.format(', '.join(names))
        )
        
    start_sync_contacts.short_description = "Sync selected characters"


@admin.register(SyncManager)
class SyncManagerAdmin(admin.ModelAdmin):
    list_display = (
        'alliance_name', 
        'contacts_count',
        'synced_characters_count', 
        'user', 
        'character_name',         
        'version_hash', 
        'last_sync',
        'last_error'
    )

    list_display_links = None

    actions = ['start_sync_managers']

    def user(self, obj):
        return obj.character.user if obj.character else None

    def character_name(self, obj):
        return obj.__str__()

    def alliance_name(self, obj):
        return obj.alliance.alliance_name

    def contacts_count(self, obj):
        return '{:,}'.format(
            obj.alliancecontact_set.count()            
        )

    def synced_characters_count(self, obj):
        return '{:,}'.format(
            obj.syncedcharacter_set.count()            
        )

    # This will help you to disbale add functionality
    def has_add_permission(self, request):
        return False

    def start_sync_managers(self, request, queryset):
                
        names = list()
        for obj in queryset:            
            tasks.run_manager_sync.delay(
                manager_pk=obj.pk, 
                force_sync=True,
                user_pk=request.user.pk
            )
            names.append(str(obj))
    
        text = 'Started syncing for: {} '.format(', '.join(names))
        text += 'You will receive a report once it is completed.'

        self.message_user(
            request, 
            text
        )
        
    start_sync_managers.short_description = "Sync selected managers"
